import numpy as np

from dp5.nmr_processing.helper_functions import generalised_lorentzian as p7


def p7residual(params, x, picked_points, y_data, region, differential):
    """
    - params: lmfit Parameters object
    """
    y = np.zeros(len(x))

    for peak in picked_points:
        y += p7(x, params['mu' + str(peak)], params['std' + str(peak)], params['vregion' + str(region)],
                params['A' + str(peak)])

    if differential == True:

        res = abs(y - y_data)

        av = np.average(res)

        av2 = np.average(res ** 2)

        difference = (res ** 2 - av2) - (res - av) ** 2

    else:
        difference = (y - y_data) ** 2

    return difference


def p7residualsolvent(params, x, picked_points, y_data, region, differential):
    y = np.zeros(len(x))

    for peak in picked_points:
        y += p7(x, params['mu' + str(peak)], params['std' + str(peak)], params['vregion' + str(region)],
                params['A' + str(peak)] * params['global_amp'])

    if differential == True:

        dy = np.gradient(y)
        ddy = np.gradient(y)

        dy_ydata = np.gradient(y_data)
        ddy_ydata = np.gradient(dy_ydata)

        difference = (y - y_data) ** 2 + (dy_ydata - dy) ** 2 + (ddy_ydata - ddy) ** 2

    else:
        difference = abs(y - y_data)

    return difference


def p7simsolvent(params, x, picked_points, region):
    y = np.zeros(len(x))

    for peak in picked_points:
        y += p7(x, params['mu' + str(peak)], params['std' + str(peak)], params['vregion' + str(region)],
                params['A' + str(peak)] * params['global_amp'])

    return y


def p7sim(params, x, picked_points, region):
    y = np.zeros(len(x))

    for peak in picked_points:
        y += p7(x, params['mu' + str(peak)], params['std' + str(peak)], params['vregion' + str(region)],
                params['A' + str(peak)])

    return y


def p7plot(params, region, group, ind, xppm):
    region_j = np.zeros(len(region))

    for peak in group:
        j = p7(region, params['mu' + str(peak)], params['std' + str(peak)], params['vregion' + str(ind)],
               params['A' + str(peak)])

        region_j += j

    return region_j


import numpy as np
from scipy.stats import norm
from matplotlib import pyplot as plt
from scipy.optimize import linear_sum_assignment as optimise
from scipy.stats import linregress
import copy
import os
import pickle

from rdkit import Chem

try:
    from openbabel.openbabel import OBConversion, OBMol, OBAtomAtomIter, OBMolAtomIter
except ImportError:
    from openbabel import *

def AssignProton(NMRData,Isomers,settings):

    #do the assignment

    for isomer in Isomers:

        assigned_shifts, assigned_peaks, assigned_labels,scaled_shifts \
            = iterative_assignment(NMRData.Hshifts,isomer.Hshifts,isomer.Hlabels,NMRData.protondata["integrals"],settings)

    #add to isomers instance

        isomer.Hexp = [''] * len(isomer.Hshifts)

        for label,peak in zip(assigned_labels,assigned_peaks):

            w = isomer.Hlabels.index(label)

            isomer.Hexp[w] = peak

    return Isomers

def iterative_assignment(exp_peaks,calculated_shifts, H_labels,rounded_integrals,settings):

    calculated_shifts = np.array(calculated_shifts)

    H_labels = np.array(H_labels)

    lnum = 0

    new_assigned_shifts = []
    old_assigned_shifts = [1]

    #print("calc shifts",calculated_shifts)

    while old_assigned_shifts != new_assigned_shifts:

        if lnum ==0:

            scaled_shifts = external_scale_proton_shifts(calculated_shifts)

            scaled_mu = 0

            scaled_std = 1

        else:
            old_assigned_shifts = copy.copy(new_assigned_shifts)
            old_assigned_peaks = copy.copy(new_assigned_peaks)

            scaled_shifts,slope,intercept = internal_scale_proton_shifts(old_assigned_shifts,old_assigned_peaks,calculated_shifts)

            scaled_std = 1

        ###############assign methyl groups first

        #find methyl groups

        m_protons = methyl_protons(settings.InputFiles[0].split('.sdf')[0] + ".sdf")

        m_shifts = np.array([])

        # find the average shifts of these groups

        for m_group in m_protons:

            s = 0

            for proton in m_group:

                w = np.where(H_labels == proton)

                s += scaled_shifts[w]/3

            m_shifts = np.hstack((m_shifts,s))

        #find peaks these can be assigned too

        methyl_peaks = []

        rounded_integrals = np.array(rounded_integrals)

        w = (rounded_integrals - (rounded_integrals % 3)) // 3

        for ind, peak in enumerate(sorted(list(set(exp_peaks)))[::-1]):
            methyl_peaks += [peak] * w[ind]

        #create difference matrix

        diff_matrix = np.zeros((len(m_shifts),len(methyl_peaks)))

        for ind1, i in enumerate(m_shifts):
            for ind2, j in enumerate(methyl_peaks):
                diff_matrix[ind1,ind2] = j-i

        prob_matrix = proton_probabilities(diff_matrix,scaled_mu,scaled_std)

        prob_matrix = prob_matrix**2

        prob_matrix = 1 - prob_matrix

        vertical_ind, horizontal_ind = optimise(prob_matrix)

        #unpack this assignment

        opt_labelsm = []

        opt_shiftsm = []

        opt_peaksm = []

        for j in vertical_ind:

            opt_labelsm.extend(m_protons[j])

        for i in horizontal_ind:

            opt_peaksm += 3*[methyl_peaks[i]]

        for label in opt_labelsm:

            w = np.where(H_labels == label)

            opt_shiftsm.append(calculated_shifts[w][0])

        #remove shifts/peaks/labels for the list to assign

        calculated_shiftsp = copy.copy(calculated_shifts)

        exp_peaksp = copy.copy(exp_peaks)

        scaled_shiftsp = copy.copy(scaled_shifts)

        H_labelsp = copy.copy(H_labels)

        #peaks

        for p in opt_peaksm:

            w = np.where(exp_peaksp == p)[0][0]

            exp_peaksp = np.delete(exp_peaksp,w)

        #shifts

        for s in opt_shiftsm:

            w = np.where(calculated_shiftsp == s)[0][0]

            calculated_shiftsp = np.delete(calculated_shiftsp,w)
            scaled_shiftsp = np.delete(scaled_shiftsp, w)

        #labels

        for l in opt_labelsm:

            w = np.where(H_labelsp == l)[0][0]

            H_labelsp = np.delete(H_labelsp,w)

        ###############assigned everything else

        diff_matrix = np.zeros((len(calculated_shiftsp),len(exp_peaksp)))

        for ind1, i in enumerate(scaled_shiftsp):
            for ind2, j in enumerate(exp_peaksp):
                diff_matrix[ind1,ind2] = j-i

        prob_matrix = proton_probabilities(diff_matrix,scaled_mu,scaled_std)

        b = abs(diff_matrix) >= 1

        ##############################find any rows that are all zeros

        b = np.where(np.sum(prob_matrix, 1) == 0)

        prob_matrix[b] =  - np.inf

        prob_matrix = np.delete(prob_matrix, b, 0)

        unassignable_shifts = calculated_shiftsp[b]

        ccalculated_shiftsp = np.delete(calculated_shiftsp, b)

        ##############################

        prob_matrix = prob_matrix**2

        prob_matrix = 1 - prob_matrix

        vertical_ind, horizontal_ind = optimise(prob_matrix)

        opt_peaksp =  exp_peaksp[horizontal_ind]

        opt_shiftsp = ccalculated_shiftsp[vertical_ind]

        opt_labelsp = H_labelsp[vertical_ind]

        opt_shifts, opt_peaks, opt_labels = removecrossassignments(opt_peaksp, opt_shiftsp, opt_labelsp)

        ################ combine these assignments

        opt_peaks = np.hstack((opt_peaksm,opt_peaksp))

        opt_shifts = np.hstack((opt_shiftsm,opt_shiftsp))

        opt_labels = np.hstack((opt_labelsm,opt_labelsp))

        #check for any shifts that have not been assigned

        copyshifts = list(copy.copy(calculated_shifts))
        copylabels = list(copy.copy(H_labels))

        for shift,label in zip(opt_shifts,opt_labels):

            copyshifts.remove(shift)
            copylabels.remove(label)

        #assign these to the closest peaks - regardless of integrals

        for shift,label in zip(copyshifts,copylabels):

            mindiff = np.array(exp_peaks - shift).argmin()

            opt_peaks = np.append(opt_peaks,exp_peaks[mindiff])

            opt_labels = np.append(opt_labels,label)

            opt_shifts = np.append(opt_shifts,shift)

        #### sort output wrt original H labels

        indv = []

        for label in opt_labels:

            wh = np.where(H_labels == label)

            indv.append(wh[0][0])

        ind = np.argsort(opt_shifts)[::-1]

        assigned_shifts = opt_shifts[indv]

        assigned_peaks = opt_peaks[indv]

        assigned_labels = opt_labels[indv]

        ind = np.argsort(assigned_shifts)

        assigned_shifts = assigned_shifts[ind].tolist()
        assigned_peaks = assigned_peaks[ind].tolist()
        assigned_labels = assigned_labels[ind].tolist()

        lnum += 1

        new_assigned_shifts =copy.copy(assigned_shifts)
        new_assigned_peaks=copy.copy(assigned_peaks)

    return assigned_shifts , assigned_peaks, assigned_labels, scaled_shifts

def external_scale_proton_shifts(calculated_shifts):

    scaled = 0.9770793502768845 * calculated_shifts - 0.019505417520415236

    return scaled

def internal_scale_proton_shifts(assigned_shifts,assigned_peaks,calculated_shifts):

    slope, intercept, r_value, p_value, std_err = linregress(assigned_shifts, assigned_peaks)

    scaled_shifts = calculated_shifts * slope + intercept

    return scaled_shifts,slope,intercept

def proton_probabilities(diff_matrix,scaled_mu,scaled_std):

    prob_matrix = norm.pdf(diff_matrix, scaled_mu, scaled_std) / norm.pdf(scaled_mu, scaled_mu, scaled_std)

    return prob_matrix

def simulate_spectrum(spectral_xdata_ppm,calc_shifts):

    y = np.zeros(len(spectral_xdata_ppm))

    for shift in calc_shifts:

        y += lorentzian(spectral_xdata_ppm,0.001,shift,0.2)

    return y

def simulate_spectrum(spectral_xdata_ppm,calc_shifts,assigned_peaks,set_exp):

    for ind,shift in enumerate(calc_shifts):

        exp_p = assigned_peaks[ind]

        ind2 = set_exp.index(exp_p)

        y = lorentzian(spectral_xdata_ppm,0.001,shift,0.2)

        plt.plot(spectral_xdata_ppm,y+1.05,color = 'C' + str(ind2 % 10))

def lorentzian(p, w, p0, A):
    x = (p0 - p) / (w / 2)
    L = A / (1 + x ** 2)

    return L

def remove_labile_protons(sdffile,lbls,shifts):

    f = sdffile.split('.sdf')[0] + '.sdf'

    mol = Chem.MolFromMolFile(f, removeHs=False)

    CI = []

    for atom in mol.GetAtoms():
        if atom.GetSymbol() == 'H':
            for neighbor in atom.GetNeighbors():
                if neighbor.GetSymbol() == 'O':
                    CI.append(f'H{atom.GetIdx()+1}')

    #remove these carbons

    for C in CI:

        ind = lbls.index(C)

        lbls.remove(C)

        for l in shifts:
            l.pop(ind)

    return lbls,shifts

def removecrossassignments(exp,calc,labels):

    #sort these in decending order

    s  = np.argsort(calc)[::-1]

    calc = calc[s]

    exp = exp[s]

    labels = labels[s]

    #generate difference matrix
    switch = 0

    expcopy = np.array(exp)

    while switch == 0:

        swapm = np.zeros([len(calc), len(calc)])

        for i,Hi in enumerate(expcopy):
            for j,Hj in enumerate(expcopy):

                if i>j:

                    swapm[i, j] = 0
                else:
                    swapm[i,j] = round(Hi - Hj,1)

        w = np.argwhere(swapm < 0)

        if len(w > 0):
            expcopy[w[0]] = expcopy[w[0][::-1]]

        else:
            switch =1

    return calc, expcopy,labels

def methyl_protons(file):

    obconversion = OBConversion()
    obconversion.SetInFormat("sdf")
    obmol = OBMol()
    obconversion.ReadFile(obmol, file)

    methyl_protons = []

    for atom in OBMolAtomIter(obmol):

        count = 0

        nbrprotons = []

        for NbrAtom in OBAtomAtomIter(atom):

            if (atom.GetAtomicNum() == 6) & (NbrAtom.GetAtomicNum() == 1):

                l = NbrAtom.GetIndex()

                count += 1

                nbrprotons.append('H' + str(l + 1))

        if count == 3:
            methyl_protons.append(nbrprotons)

    return methyl_protons



import re
import networkx as nx
import logging

logger = logging.getLogger(__name__)


def process_description(nmr_source):
    with open(nmr_source) as f:
        Cexp = f.readline()
        f.readline()
        Hexp = f.readline()
        equivalents = []
        omits = []
        f.readline()

        for line in f:
            if not "OMIT" in line and len(line) > 1:
                equivalents.append(line[:-1].split(","))
            elif "OMIT" in line:
                omits.extend(line[5:-1].split(","))
    logger.info("Read carbon NMR shifts")
    C_labels, C_exp = _parse_description(Cexp)
    logger.info("Read proton NMR shifts")
    H_labels, H_exp = _parse_description(Hexp)

    return C_labels, C_exp, H_labels, H_exp, equivalents, omits


def _parse_description(exp):

    if len(exp) > 0:
        # Replace all 'or' and 'OR' with ',', remove all spaces and 'any'
        texp = re.sub(r"or|OR", ",", exp, flags=re.DOTALL)
        texp = re.sub(r" ", "", texp, flags=re.DOTALL)
        # Get all assignments, split mulitassignments
        expLabels = re.findall(r"(?<=\().*?(?=\)|;)", texp, flags=re.DOTALL)
        expLabels = [x.replace("any", "") for x in expLabels]
        expLabels = [x.split(",") for x in expLabels]
        # Remove assignments and get shifts
        ShiftData = (re.sub(r"\(.*?\)", "", exp.strip(), flags=re.DOTALL)).split(",")
        logger.info(", ".join(ShiftData))
        expShifts = [float(x) for x in ShiftData]
    else:
        expLabels = []
        expShifts = []

    return expLabels, expShifts


def pairwise_assignment(calculated, experimental: list):
    sorted_calc = sorted(calculated, reverse=True)
    sorted_exp = sorted(experimental, reverse=True)
    assigned = [None] * len(calculated)

    for calc, exp in zip(sorted_calc, sorted_exp):
        index = list(calculated).index(calc)
        assigned[index] = exp

    return assigned


def matching_assignment(calculated, experimental, threshold=40):

    scaled = calculated

    # Create a bipartite graph
    G = nx.Graph()

    # Add nodes for calc and exp with a bipartite attribute
    calc_nodes = [("calc", i) for i in range(len(scaled))]
    exp_nodes = [("exp", i) for i in range(len(experimental))]
    G.add_nodes_from(calc_nodes, bipartite=0)
    G.add_nodes_from(exp_nodes, bipartite=1)

    # Add edges for all pairs within the threshold
    for i, c in enumerate(scaled):
        for j, e in enumerate(experimental):
            deviation = abs(c - e)
            if deviation <= threshold:
                G.add_edge(("calc", i), ("exp", j), weight=-deviation)

    # Find the maximum matching
    matching = nx.algorithms.matching.max_weight_matching(G, maxcardinality=True)

    matched_pair_indices = set()

    for pair in matching:
        # pair could be (('calc', i), ('exp', j)) or the reverse
        calc, exp = sorted(pair)
        # after sorting, 'calc' goes before 'exp'
        calc_index = calc[1]
        exp_index = exp[1]

        matched_pair_indices.add((calc_index, exp_index))

    # Now, use these indices to access values in calc and exp
    matched_pairs = [(calculated[i], experimental[j]) for i, j in matched_pair_indices]
    assigned = [None] * len(calculated)
    for calc_shift, exp_shift in matched_pairs:
        index = list(calculated).index(calc_shift)
        assigned[index] = exp_shift

    return assigned



from rdkit import Chem
from rdkit.Chem import AllChem
import os
from InchiGen import *

def GenerateMolFromSDF(InputFile):

    cwd = os.getcwd()

    fullf = os.path.join(cwd, InputFile + '.sdf')

    m = Chem.MolFromMolFile(fullf, removeHs=False,sanitize = True)

    return m

def GenerateSDFFromTxt(InputFile,inp_type):

    Mols = []

    cwd = os.getcwd()

    InputFile

    if os.path.exists(os.path.join(cwd, InputFile)):

        fullf = os.path.join(cwd, InputFile)

    elif inp_type == 'InChI':

        fullf = os.path.join(cwd, InputFile + ".inchi")

    elif inp_type == 'Smiles':

        fullf = os.path.join(cwd, InputFile + ".smiles")

    elif inp_type == 'Smarts':

        fullf = os.path.join(cwd, InputFile + ".smarts")

    f = open(fullf, "r")

    if inp_type == 'InChI':

        for line in f.readlines():

            if line.strip() == '':
                continue

            else:

                try:

                    m = Chem.MolFromInchi(line.strip(),sanitize = True)

                    Mols.append(m)

                except:

                    print(line + " could not be read")

    elif inp_type == 'Smiles':

        for line in f.readlines():

            if line.strip() == '':
                continue

            else:

                try:

                    m = Chem.MolFromSmiles(line.strip(),sanitize = True)

                    Mols.append(m)

                except:

                    print(line + " could not be read")

    elif inp_type == 'Smarts':

        for line in f.readlines():

            if line.strip() == '':
                continue

            else:

                try:

                    m = Chem.MolFromSmarts(line.strip(),sanitize = True)

                    Mols.append(m)

                except:

                    print(line + " could not be read")

    else:

        print('unrecognised')

    GeneratedFiles = GenerateSDFFromMols(Mols,inp_type )

    return GeneratedFiles

def GenerateSDFFromMols(mols,inp_type):

    files = []

    cwd = os.getcwd()

    for ind, m in enumerate(mols):

        m = AllChem.AddHs(m, addCoords=True)

        AllChem.EmbedMolecule(m)

        AllChem.MMFFOptimizeMolecule(m)

        Chem.rdmolops.AssignStereochemistryFrom3D(m)

        f = inp_type +  '_Mol_' + str(ind) +"_.sdf"

        files.append(f[:-4])

        fullf = os.path.join(cwd, f)

        save3d = Chem.SDWriter( fullf)

        save3d.write(m)

    return files

def CleanUp(InputFiles):

    # check input file types

    CleanedInputFiles = []

    cwd = os.getcwd()

    for f in InputFiles:

        if f.endswith('.sdf'):

            f = f[:-4]

        fullf = os.path.join(cwd, f + 'cleaned.sdf')

        m = GenerateMolFromSDF(f)

        m = AllChem.AddHs(m, addCoords=True)

        AllChem.EmbedMolecule(m)

        AllChem.MMFFOptimizeMolecule(m)

        Chem.rdmolops.AssignStereochemistryFrom3D(m)

        save3d = Chem.SDWriter(fullf)

        save3d.write(m)

        CleanedInputFiles.append( f + 'cleaned')

    return CleanedInputFiles

def NumberofStereoCentres(InputFile):

    cwd = os.getcwd()

    fullf = os.path.join(cwd, InputFile + '.sdf')

    m = Chem.MolFromMolFile(fullf, removeHs=False)

    m = AllChem.AddHs(m, addCoords=True)

    AllChem.EmbedMolecule(m)

    AllChem.MMFFOptimizeMolecule(m)

    Chem.rdmolops.AssignStereochemistryFrom3D(m)

    nStereo = Chem.rdMolDescriptors.CalcNumAtomStereoCenters(m)

    return nStereo


import copy

import numpy as np

from dp5.nmr_processing.proton.fid_corrections import baseline_find_signal


def gradient_peak_picking(y_data, corr_distance, uc, std, binary_map_regions):
    final_peaks = []

    # estimate std of second derivative data

    ddy = np.diff(y_data, 2)

    ddy = ddy / np.max(ddy)

    # find peaks

    classification, sigma = baseline_find_signal(
        -1 * ddy, corr_distance, False, 2)

    ddy1 = np.roll(ddy, 1)

    ddyn1 = np.roll(ddy, -1)

    p = np.where((ddy < ddy1) & (ddy < ddyn1))[0]

    peaks = p[classification[p] == 1]

    peaks1 = np.roll(peaks, 1)

    distance = np.min(abs(peaks1 - peaks))

    # must make sure the convolution kernel is odd in length to prevent the movement of the peaks

    peaks = np.sort(peaks)

    peakscopy = copy.copy(peaks)
    ddycopy = copy.copy(ddy[peaks] / np.max(ddy))

    while distance < corr_distance:
        # roll the peaks one forward
        peakscopy1 = np.roll(peakscopy, 1)

        # find distances between peaks
        diff = np.abs(peakscopy - peakscopy1)

        # find where in the array the smallest distance is
        mindist = np.argmin(diff)

        # what is this distance
        distance = diff[mindist]

        # compare the values of the second derivative at the closest two peaks

        compare = np.argmax(ddycopy[[mindist, mindist - 1]])

        peakscopy = np.delete(peakscopy, mindist - compare)
        ddycopy = np.delete(ddycopy, mindist - compare)

    # remove any peaks that fall into the noise

    n = y_data[peakscopy]

    w = n > 5 * std

    peakscopy = peakscopy[w]

    final_peaks = np.sort(peakscopy)

    # draw new regions symmetrically around the newly found peaks
    # introduces assumption of J not exceeding 18 Hz

    dist_hz = uc(0, "Hz") - uc(9, "Hz")

    peak_regions = np.empty(shape=final_peaks.shape[0], dtype=object)

    for i, peak in enumerate(final_peaks):

        right = min(peak + dist_hz + 1, len(y_data))
        left = max(peak - dist_hz, 0)
        region = np.arange(left, right, 1)

        peak_regions[i] = region

    final_regions = [peak_regions[0]]
    final_peaks_separated = [[final_peaks[0]]]
    # logic: if regions overlap, stitch together
    for region in range(1, len(peak_regions)):

        if peak_regions[region][0] <= final_regions[-1][-1]:

            final_regions[-1] = np.union1d(final_regions[-1],
                                           peak_regions[region])

            final_peaks_separated[-1].append(final_peaks[region])

        else:

            final_regions += [np.array(peak_regions[region])]

            final_peaks_separated.append([final_peaks[region]])
    # clean up overlapping regions
    # consistent numpyfication for subsequent operations
    final_regions = np.asarray(final_regions, dtype=object)
    final_peaks_separated = [np.array(i) for i in final_peaks_separated]
    final_peaks_separated = np.asarray(final_peaks_separated, dtype=object)

    return final_peaks, final_regions, final_peaks_separated, std



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on Tue Mar 31 14:54:17 2015

@author: ke291
"""
import sys

def FindAllPaths(molgraph, start, end, path=[]):
    path = path + [start]
    if start == end:
        return [path]
    if start not in [x[0] for x in molgraph]:
        print("No such node in graph")
        return []
    paths = []
    for node in molgraph[start-1][1:]:
        if node not in path:
            newpaths = FindAllPaths(molgraph, node, end, path)
            for newpath in newpaths:
                paths.append(newpath)
    return paths


def FindTerminatingPaths(molgraph, start, trunk, path=[]):
    path = path + [start]
    if start not in [x[0] for x in molgraph]:
        print("No such node in graph")
        return []
    paths = []
    for node in molgraph[start-1][1:]:
        if node not in path and node not in trunk:
            newpaths = FindTerminatingPaths(molgraph, node, trunk, path)
            for newpath in newpaths:
                paths.append(newpath)
    if paths == []:
        return [path]
    else:
        return paths


def FindTreeMap(f):
    #get molecular graph
    molgraph = GenMolGraph(f)

    #Find potential endpoints - atoms with only one neighbour
    endpoints = []
    for node in molgraph:
        if len(node) < 3:
            endpoints.append(node[0])

    #get the longest paths for all endpoint combinations
    maxpaths = []
    for i in range(0, len(endpoints)):
        for j in range(0, len(endpoints)):
            if i != j:
                maxpaths.append(max(FindAllPaths(molgraph, endpoints[i],
                                                 endpoints[j]), key=len))
    #get the absolute longest path possible in the molecule
    molmap = max(maxpaths, key=len)
    #Add longest branches to the longest trunk
    for atom in molmap:
        for node in molgraph[atom-1]:
            if node not in molmap:
                maxbranch = []
                branches = FindTerminatingPaths(molgraph, node, molmap)
                if branches != []:
                    maxbranch = max(branches, key=len)
                if maxbranch != []:
                    molmap.extend(maxbranch)
    return molmap


def TreeRenumSDF(f, ExpNMR):

    molmap = FindTreeMap(f)

    #Read molecule from file
    obconversion = OBConversion()
    obconversion.SetInFormat("sdf")
    obmol = OBMol()
    obconversion.ReadFile(obmol, f)

    temp = []
    anums = []
    for atom in OBMolAtomIter(obmol):
        temp.append(atom)
        anums.append(atom.GetAtomicNum())

    newmol = OBMol()
    for atom in molmap:
        newmol.AddAtom(temp[atom-1])

    #Restore the bonds
    newmol.ConnectTheDots()
    newmol.PerceiveBondOrders()
    #Write renumbered molecule to file
    obconversion.SetOutFormat("sdf")
    obconversion.WriteFile(newmol, f[:-4] + 'r.sdf')

    #Prepare NMR translation
    NMRmap = []
    i = 1
    for atom in molmap:
        anum = anums[atom-1]
        if anum == 1:
            NMRmap.append(['H' + str(atom), 'H' + str(i)])
        if anum == 6:
            NMRmap.append(['C' + str(atom), 'C' + str(i)])
        i += 1
    print(NMRmap)
    RenumNMR(ExpNMR, NMRmap)


def RenumNMR(ExpNMR, NMRmap):
    f = open(ExpNMR, 'r')
    NMRfile = f.read(1000000)
    f.close()

    print('\nOld NMR file:\n' + NMRfile)

    #Replace old atom labels with new atom labels
    #tag replacements with '_' to avoid double replacement
    for atom in NMRmap:
        NMRfile = NMRfile.replace(atom[0] + ')', atom[1] + '_)')
        NMRfile = NMRfile.replace(atom[0] + ' ', atom[1] + '_ ')
        NMRfile = NMRfile.replace(atom[0] + ',', atom[1] + '_,')
        NMRfile = NMRfile.replace(atom[0] + '\n', atom[1] + '_\n')

    #Strip temporary udnerscore tags
    NMRfile = NMRfile.replace('_', '')

    print('\nNew NMR file:\n' + NMRfile)
    f = open(ExpNMR + 'r', 'w')
    f.write(NMRfile)
    f.close()


def GenMolGraph(f):
    obconversion = OBConversion()
    obconversion.SetInFormat("sdf")
    obmol = OBMol()
    obconversion.ReadFile(obmol, f)

    molgraph = []

    for atom in OBMolAtomIter(obmol):
        idx = atom.GetIdx()
        molgraph.append([])
        molgraph[idx-1].append(idx)

        for NbrAtom in OBAtomAtomIter(atom):
            molgraph[idx-1].append(NbrAtom.GetIdx())

    return molgraph

if __name__ == '__main__':
    #print sys.argv
    ExpNMR = sys.argv[2]
    filename = sys.argv[1] + '.sdf'
    TreeRenumSDF(filename, ExpNMR)
    #main()


"""
wrapper for DFT methods.

Should implement setup, running, and reading the calculations
"""

import time
import subprocess
import sys
import shutil
from abc import abstractmethod, ABC
from math import isnan
import logging
from pathlib import Path


logger = logging.getLogger(__name__)


class BaseDFTMethod(ABC):

    hartree_to_kJ = 2625.499629554010
    gasConstant = 8.3145
    temperature = 298.15

    def __init__(self, settings):
        self.settings = settings
        if not isnan(self.settings["charge"]):
            self.charge = self.settings["charge"]
        else:
            self.charge = None
        self.force_converge = self.settings["optimisation_converged"]
        self.dft_complete = self.settings["dft_complete"]
        self.memory = self.settings["memory"]
        self.num_processors = self.settings["num_processors"]

        self.tag = ""
        self.input_format = ".in"
        self.output_format = ".out"
        self.executable = "echo"

    @abstractmethod
    def __repr__(self) -> str:
        return "DFTMethod"

    def _get_files(self, mols, calc_type):
        jobdir = Path.cwd()

        data = jobdir / calc_type
        data.mkdir(exist_ok=True)

        files_to_run = []
        input_files = {mol.base_name: [] for mol in mols}
        output_files = {mol.base_name: [] for mol in mols}

        for mol in mols:
            if self.charge:
                charge = self.charge
            else:
                charge = mol.charge

            for i, conf in enumerate(mol.conformers, start=1):
                filename = data / f"{mol}{self.tag}inp{i:03}"
                input_name = filename.with_suffix(self.input_format)
                output_name = filename.with_suffix(self.output_format)
                geom = conf

                if output_name.exists():
                    logger.debug(f"Reading DFT data from file {output_name}")
                    atoms, coords, *_, completed, converged = self.read_file(
                        output_name
                    )
                    if completed:
                        if calc_type == "opt":
                            if converged or self.force_converge:
                                output_files[mol.base_name].append(filename)
                                continue
                            else:
                                if coords:
                                    logger.info(
                                        f"Reusing partially optimised coordinates for {filename}"
                                    )
                                    geom = coords
                                output_name.unlink()
                        else:
                            output_files[mol.base_name].append(filename)
                            continue
                    else:
                        output_name.unlink()
                logger.debug(f"Creating new DFT file {input_name}")
                self.write_file(filename, geom, mol.atoms, charge, calc_type)
                input_files[mol.base_name].append(filename)
                files_to_run.append(filename)

        if files_to_run:
            logger.debug(f"Starting DFT calculations for {calc_type}")
            completed = self._run_calcs(files_to_run)

            for mol in mols:
                for file in completed:
                    if file in input_files[mol.base_name]:
                        output_files[mol.base_name].append(file)

        output_files = [output_files[key] for key in sorted(output_files.keys())]

        return output_files

    def _get_prerun_files(self, mols, calc_type: str):
        """
        Gathers and returns output files for calculations.
        Assumes all your files have terminated normally.

        Arguments:
        - mols: a list of Molecule objects
        - calc_type: a type of calculation.
        Returns:
         - list of lists of output files
        """
        jobdir = Path.cwd()
        data = jobdir / calc_type

        prerun_files = []

        if data.is_dir():
            for mol in mols:
                outputs = []
                pattern = f"{mol}{self.tag}inp*{self.input_format}"
                logger.debug(f"Searching for {pattern}")
                inputs = sorted([file for file in data.glob(pattern)])
                for input_file in inputs:
                    output_file = input_file.with_suffix(self.output_format)
                    if not output_file.exists():
                        logger.warning(f"Cannot find output file for {input_file}")
                        continue
                    if self.is_completed(output_file):
                        outputs.append(output_file)
                    else:
                        logger.warning(f"Calculations not complete for {output_file}")
                prerun_files.append(outputs)
        else:
            logger.error(f"No folder found at {data}!")

        return prerun_files

    def get_files(self, mols, calc_type):
        if self.dft_complete:
            logger.debug("Loading pre-run files")
            files = self._get_prerun_files(mols, calc_type)
        else:
            logger.debug("No pre-run files found, starting new calculations")
            files = self._get_files(mols, calc_type)
        return files

    def opt(self, mols):
        files = self.get_files(mols, "opt")

        atoms = []
        conformers = []
        energies = []

        for mol_data in files:
            mol_conformers = []
            mol_energies = []
            for file in mol_data:
                filename = file.with_suffix(self.output_format)
                logger.debug(f"Reading DFT output file: {filename}")
                mol_atoms, coords, energy, *_, opt_converged = self.read_file(filename)

                if opt_converged or self.force_converge:
                    mol_conformers.append(coords)
                    mol_energies.append(energy)
                else:
                    logger.critical(
                        f"Optimisation for {file} has not converged! Terminating..."
                    )
                    sys.exit(1)

            atoms.append(mol_atoms)
            conformers.append(mol_conformers)
            energies.append(mol_energies)

        return atoms, conformers, energies

    def energy(self, mols):
        files = self.get_files(mols, "e")

        atoms = []
        conformers = []
        energies = []

        for mol_data in files:
            mol_conformers = []
            mol_energies = []
            for file in mol_data:
                filename = file.with_suffix(self.output_format)
                logger.debug(f"Reading DFT output file: {filename}")
                mol_atoms, coords, energy, *_ = self.read_file(filename)
                mol_conformers.append(coords)
                mol_energies.append(energy)

            atoms.append(mol_atoms)
            conformers.append(mol_conformers)
            energies.append(mol_energies)

        return atoms, conformers, energies

    def nmr(self, mols):
        files = self.get_files(mols, "nmr")

        atoms = []
        conformers = []
        energies = []
        shieldings = []
        shielding_labels = []

        for mol_data in files:
            mol_conformers = []
            mol_energies = []
            mol_shieldings = []
            mol_shielding_labels = []
            for file in mol_data:
                filename = file.with_suffix(self.output_format)
                logger.debug(f"Reading DFT output file: {filename}")
                mol_atoms, coords, energy, shielding, shielding_label, *flags = self.read_file(filename)
                mol_conformers.append(coords)
                mol_energies.append(energy)
                mol_shieldings.append(shielding)
                mol_shielding_labels.append(shielding_label)

            atoms.append(mol_atoms)
            conformers.append(mol_conformers)
            energies.append(mol_energies)
            shieldings.append(mol_shieldings)
            shielding_labels.append(mol_shielding_labels)

        return atoms, conformers, energies, shieldings, shielding_labels

    @abstractmethod
    def write_file(self, filename, coordinates, atoms, charge, calc_type):
        raise NotImplementedError("not implemented")

    def _run_calcs(self, jobs: list):
        """
        - jobs: list of gaussian input files for which calculations are required
        """
        num_complete = 0
        jobs_complete = []

        if shutil.which(self.executable) is None:
            logger.critical(
                f"{self} executable not found at {self.executable}. Terminating..."
            )
            sys.exit(1)

        for file in jobs:
            logger.info(f"Starting calculations for {file}")
            cmd = self.prepare_command(file)
            outp = subprocess.check_output(cmd, shell=True, timeout=86400)

            num_complete += 1
            if self.is_completed(file.with_suffix(self.output_format)):
                jobs_complete.append(file)
                logger.info(f"{self} job {num_complete} of {len(jobs)} completed")
            else:
                logger.info(f"Job {file} terminated with an error. Continuing")

        if num_complete > 0:
            logger.info(f"{num_complete} jobs completed successfully")

        return jobs_complete

    @abstractmethod
    def prepare_command(self, file):
        return (
            f"{self.executable} {file}{self.input_format} > {file}{self.output_format}"
        )

    def is_converged(self, file):
        *_, converged = self.read_file(file)
        return converged

    def is_completed(self, file):
        *_, completed, converged = self.read_file(file)
        return completed

    @abstractmethod
    def read_file(self, file):
        """Reads output file. Inspired by cclib parser

        Returns:
            - atoms
            - coordinates
            - energies
            - shieldings
            - shielding labels
            - if calculation terminated normally
            - if optimisation has converged

        """
        atoms = []
        coordinates = []
        energy = None
        shieldings = []
        shielding_labels = []
        # check for normal termination
        completed = False
        # check for optimisation convergence
        converged = False
        with open(file, "r") as f:
            while not completed:
                try:
                    line = next(f)  # get new line using next
                    if "first line of geometry block":
                        atoms.append("atom")
                        coordinates.append("coordinates")

                    if "shielding":
                        shieldings.append("shielding")
                        shielding_labels.append("shielding label")

                    if "optimisation converged":
                        converged = True

                    if "energy":
                        energy = 42

                    if "normal termination":
                        completed = True
                    # add your geometry section, compare with gaussian.py!
                except StopIteration:
                    logger.error(
                        f"Calculations in {file} have not terminated normally."
                    )
                    break
            # write your functions
        return (
            atoms,
            coordinates,
            energy,
            shieldings,
            shielding_labels,
            completed,
            converged,
        )

    def atom_num_to_symbol(self, anum: int) -> str:

        Lookup = [
            "H",
            "He",
            "Li",
            "Be",
            "B",
            "C",
            "N",
            "O",
            "F",
            "Ne",
            "Na",
            "Mg",
            "Al",
            "Si",
            "P",
            "S",
            "Cl",
            "Ar",
            "K",
            "Ca",
            "Sc",
            "Ti",
            "V",
            "Cr",
            "Mn",
            "Fe",
            "Co",
            "Ni",
            "Cu",
            "Zn",
            "Ga",
            "Ge",
            "As",
            "Se",
            "Br",
            "Kr",
            "Rb",
            "Sr",
            "Y",
            "Zr",
            "Nb",
            "Mo",
            "Tc",
            "Ru",
            "Rh",
            "Pd",
            "Ag",
            "Cd",
            "In",
            "Sn",
            "Sb",
            "Te",
            "I",
            "Xe",
            "Cs",
            "Ba",
            "La",
            "Ce",
            "Pr",
            "Nd",
            "Pm",
            "Sm",
            "Eu",
            "Gd",
            "Tb",
            "Dy",
            "Ho",
            "Er",
            "Tm",
            "Yb",
            "Lu",
            "Hf",
            "Ta",
            "W",
            "Re",
            "Os",
            "Ir",
            "Pt",
            "Au",
            "Hg",
            "Tl",
            "Pb",
            "Bi",
            "Po",
            "At",
            "Rn",
        ]

        if anum > 0 and anum < len(Lookup):
            return Lookup[anum - 1]
        else:
            print(f"No such element with atomic number {anum}")
            return 0



import logging
from typing import List
import sys
from importlib import import_module
from pathlib import Path
import dataclasses

from dp5.conformer_search import conf_prune

logger = logging.getLogger(__name__)


def conf_search(mols, config) -> List:
    """
    arguments:
    - mols: list of Molecule objects, handled by the Molecules container
    - config: conformational search config
    """

    try:
        Method = import_module(f'.{config["method"]}', "dp5.conformer_search")
        logger.debug(f"Loaded {config['method']} as conformational search method")
    except ModuleNotFoundError:
        logger.critical(f"No conformer search module {config['method']} found")
        sys.exit(1)
        raise ValueError(f"No conformer search module {config['method']} was found")

    try:
        method = Method.ConfSearchMethod(config)
        logger.debug(f"Initialised {config['method']}")
    except AttributeError:
        logger.critical(
            f"{config['method']} does not implement ConfSearchMethod. Terminating."
        )
        sys.exit(1)
        raise ValueError(f"{config['method']} does not implement ConfSearchMethod.")
    # input names have no extension!
    input_names = [mol.base_name for mol in mols]
    final_names = input_names.copy()

    flipped_inputs = []
    if config["manual_five_membered_rings"]:
        logger.info("Using manual procedure to flip five-membered rings")
        import five_conf

        for file in input_names:
            if not Path(f"{file}rot.sdf").exists():
                # returns empty string if unnecessary
                rot = five_conf.main(f"{file}.sdf", config["five_membered_ring_atoms"])
                if not rot:
                    continue
            flipped_inputs.append(f"{file}rot")
        final_names.extend(flipped_inputs)

    logger.info(f"Using {config['method']} for conformational search")
    cs_data = method(final_names)
    logger.info("Conformational search complete")

    merged_data = []
    logger.info(f"Reading output files: {input_names}")
    for mol, data in zip(input_names, cs_data):
        coords = data.conformers
        energies = data.energies
        logger.debug(f"{len(coords)} conformers read for {mol}")

        if f"{mol}rot" in flipped_inputs:
            # merges the rotated inputs
            logger.info(f"found {mol}rot, appending to {mol}")
            rot_idx = final_names.index(f"{mol}rot")
            rot_data = cs_data[rot_idx]
            logger.debug(
                f"{len(rot_data.conformers)} rotated conformers read for {mol}rot"
            )
            coords.extend(rot_data.conformers)
            coords.extend(rot_data.energies)

        all_conformers = [(e, c) for e, c in zip(energies, coords)]
        all_conformers.sort(key=lambda x: x[0])

        try:
            emin = all_conformers[0][0]
        except:
            logger.error("Cannot find minimum energy value")
            logger.error(f"The file {mol} was not parsed correctly")
            raise Exception(f"The file {mol} was not parsed correctly")

        all_conformers = [
            (energy, coordinates)
            for energy, coordinates in all_conformers
            if energy < emin + config["energy_cutoff"]
        ]
        logger.debug(
            f"{len(all_conformers)} total conformers read within {config['energy_cutoff']:.1f} kJ/mol for {mol}"
        )

        if len(all_conformers) > config["conf_per_structure"] and config["conf_prune"]:
            logger.info(f"Conformer pruning required for {mol}")
            unique_conf_ids, cutoff = conf_prune(
                conformers=coords,
                cutoff=config["rmsd_cutoff"],
                conf_limit=config["conf_per_structure"],
            )

            all_conformers = [
                conformer
                for conformer, unique in zip(all_conformers, unique_conf_ids)
                if unique
            ]
            logger.info(f"Retained {len(all_conformers)} conformers for {mol}")
            logger.info(f"Adjusted RMSD threshold: {cutoff:.2f} \u00c5")

        energies, coords = zip(*all_conformers)

        data = dataclasses.replace(data, conformers=coords, energies=energies)

        merged_data.append(data)

    return merged_data



class Tokenizer(object):
    """ A class to turn arbitrary inputs into integer classes. """

    def __init__(self):
        # the default class for an unseen entry during test-time
        self._data = {'unk': 1}
        self.num_classes = 1
        self.train = True
        self.unknown = []

    def __call__(self, item):
        """ Check to see if the Tokenizer has seen `item` before, and if so,
        return the integer class associated with it. Otherwise, if we're
        training, create a new integer class, otherwise return the 'unknown'
        class.

        """
        try:
            return self._data[item]

        except KeyError:
            if self.train:
                self._add_token(item)
                return self(item)

            else:
                # Record the unknown item, then return the unknown label
                self.unknown += [item]
                return self._data['unk']

    def _add_token(self, item):
        self.num_classes += 1
        self._data[item] = self.num_classes

# The rest of the methods in this module are specific functions for computing
# atom and bond features. New ones can be easily added though, and these are
# passed directly to the Preprocessor class.

def get_ring_size(obj, max_size=12):
    if not obj.IsInRing():
        return 0
    else:
        for i in range(max_size):
            if obj.IsInRingSize(i):
                return i
        else:
            return 'max'

def atom_features(atom):
    return atom.GetAtomicNum()


def atom_features_v1(atom):
    """ Return an integer hash representing the atom type
    """

    return str((
        atom.GetSymbol(),
        atom.GetDegree(),
        atom.GetTotalNumHs(),
        atom.GetImplicitValence(),
        atom.GetIsAromatic(),
    ))


def atom_features_v2(atom):

    props = ['GetChiralTag', 'GetDegree', 'GetExplicitValence',
             'GetFormalCharge', 'GetHybridization', 'GetImplicitValence',
             'GetIsAromatic', 'GetNoImplicit', 'GetNumExplicitHs',
             'GetNumImplicitHs', 'GetNumRadicalElectrons', 'GetSymbol',
             'GetTotalDegree', 'GetTotalNumHs', 'GetTotalValence']

    atom_type = [getattr(atom, prop)() for prop in props]
    atom_type += [get_ring_size(atom)]

    return str(tuple(atom_type))


def bond_features_v1(bond, **kwargs):
    """ Return an integer hash representing the bond type.

    flipped : bool
        Only valid for 'v3' version, whether to swap the begin and end atom types

    """

    return str((
        bond.GetBondType(),
        bond.GetIsConjugated(),
        bond.IsInRing(),
        sorted([
            bond.GetBeginAtom().GetSymbol(),
            bond.GetEndAtom().GetSymbol()]),
        ))


def bond_features_v2(bond, **kwargs):

    return str((
        bond.GetBondType(),
        bond.GetIsConjugated(),
        bond.GetStereo(),
        get_ring_size(bond),
        sorted([
            bond.GetBeginAtom().GetSymbol(),
            bond.GetEndAtom().GetSymbol()]),
        ))


def bond_features_v3(bond, flipped=False):

    if not flipped:
        start_atom = atom_features(bond.GetBeginAtom())
        end_atom = atom_features(bond.GetEndAtom())

    else:
        start_atom = atom_features(bond.GetEndAtom())
        end_atom = atom_features(bond.GetBeginAtom())

    return str((
        bond.GetBondType(),
        bond.GetIsConjugated(),
        bond.GetStereo(),
        get_ring_size(bond),
        bond.GetEndAtom().GetSymbol(),
        start_atom,
        end_atom))



from dp5.dft.gaussian import DFTMethod
from dp5.run.data_structures import get
from dp5.run.data_structures import Molecule
from dp5.nmr_processing.nmr_ai import NMRData
import os

s11 = '/home/rk582/DP5/S11'

nmr_data = NMRData(['/home/rk582/DP5/S11/S11_NMR'], 'chloroform')

mol = Molecule('S11/S11b_.sdf')

settings = {'charge': 0, 'optimisation_converged': True,
            'dft_complete': True, 'memory': 1000, 'num_processors': 42, 'executable': {'gaussian': ''}}

os.chdir('/home/rk582/DP5/S11')
g16 = DFTMethod(settings)
g16.nmr([mol])



"""
Sets up a template for conformer search class. Uses run as main method.
Should return a list of atoms, conformer geometries, charges, and energies.
"""

from abc import abstractmethod, ABC
from dataclasses import dataclass
import logging
from typing import List


logger = logging.getLogger(__name__)


class BaseConfSearch(ABC):

    def __init__(self, settings):
        self.settings = settings

    @abstractmethod
    def __repr__(self) -> str:
        raise NotImplementedError("No __repr__ attribute provided.")

    def __call__(self, inputs) -> list:
        """
        Handles input preparation, execution, and processing. The only method called externally.

        """
        logger.info(f"Using {self} as conformer search method")
        self.inputs = self.prepare_input(inputs)
        self.outputs = self._run()
        logger.debug(f"Conformer search output: {self.outputs}")

        return self.parse_output()

    @abstractmethod
    def _run(self):
        """
        Executes the conformer search.
        Arguments:
        - self.inputs: specified at initialisation
        Returns:
        - paths to output files
        """
        raise NotImplementedError("run method is not implemented")

    @abstractmethod
    def prepare_input(self):
        raise NotImplementedError("prepare method is not implemented")

    @abstractmethod
    def _parse_output(self, file):
        raise NotImplementedError("Method for reading output files is not implemented")

    def parse_output(self):
        final_list = []
        for file in self.inputs:
            conf_data = self._parse_output(file)
            assert isinstance(
                conf_data, ConfData
            ), "Please pack your conformer search data into ConfData"
            final_list.append(conf_data)

        return final_list


@dataclass(frozen=True)
class ConfData:
    atoms: List[str]
    conformers: List[List[List[float]]]
    charge: int
    energies: List[float]



from keras.models import Model

class GraphModel(Model):
    """ This is a simple modification of the Keras `Model` class to avoid
    checking each input for a consistent batch_size dimension. Should work as
    of keras-team/keras#11548.

    """

    def _standardize_user_data(self, *args, **kwargs):
        kwargs['check_array_lengths'] = False
        return super(GraphModel, self)._standardize_user_data(*args, **kwargs)



import pickle
from dp5.run.data_structures import Molecule
from dp5.nmr_processing.proton import integrate
from rdkit import Chem
from rdkit.Chem import AllChem
import os

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'


mol = Chem.MolFromMolFile('/home/rk582/DP5/test/tegan_mosher.mol')
mol_h = Chem.AddHs(mol, addCoords=True)


with open('/home/rk582/DP5/test/tps-1-14.preint', 'rb') as f:
    data = pickle.load(f)

# currently crashes in integrate_sim_regions - verify correct arguments are passed
# fails due to lack of grouping???? check remove_impurities
# integration is odd, numpy refuses to filter peaks
result = integrate.find_integrals(mol_h, data.protondata["peak_regions"],
                                  data.protondata["grouped_peaks"],
                                  data.protondata["sim_regions"],
                                  data.protondata["picked_peaks"],
                                  data.protondata["params"],
                                  data.protondata["ydata"],)

_ = result



import numpy as np
import pickle
from scipy.stats import gmean
from pathlib import Path
from scipy import stats
import os
import pathos.multiprocessing as mp
from rdkit import Chem
from rdkit.Geometry import Point3D
from rdkit.Chem import AllChem
try:
    from openbabel.openbabel import OBConversion, OBMol, OBAtomAtomIter, OBMolAtomIter
except ImportError:
    from openbabel import *

import pandas as pd
import dp5.neural_net.CNN_model as CNN_model

c_distance = 4.532297920317418

class DP5data:

    def __init__(self,ScriptPath,Atoms):

        self.Atom_number = Atoms

        self.Cshifts = []  # Carbon shifts used in DP5 calculation
        self.Cscaled = []  # Scaled Carbon shifts used in DP5 calculation
        self.Cexp = []  # Carbon experimental shifts used in DP4 calculation
        self.Clabels = []  # Carbon atom labels
        self.Cinds = [] # Carbon atom indcies
        self.Hshifts = []  # Proton shifts used in DP5 calculation
        self.Hscaled = []  # Scaled Proton shifts used in DP5 calculation
        self.Hexp = []  # Proton experimental shifts used in DP5 calculation
        self.Hlabels = []  # Proton atom labels
        self.CMAE = [] # Carbon Mean Absolute error
        self.CMax = [] # Carbon Max Absolute error
        self.ConfCshifts = []

        self.Mols = [] #qml compound objects for each isomer

        self.ErrorAtomReps = [] # FCHL representations ordered by Clabels

        self.ErrorAtomProbs = [] #per atom dp5 scaled probabilities for all conformers

        self.B_ErrorAtomProbs = [] #per atom dp5 Exp probabilities boltzmann weighted

        self.Mol_Error_probs = []  # DP5 for isomers based on Carbon data

        self.DP5_Error_probs = []  # Final DP5S

        self.ExpAtomReps = [] # FCHL representations ordered by Clabels

        self.ExpAtomProbs = [] #per atom dp5 scaled probabilities for all conformers

        self.B_ExpAtomProbs = [] #per atom dp5 Exp probabilities boltzmann weighted

        self.Mol_Exp_probs = []  # DP5 for isomers based on Carbon data

        self.DP5_Exp_probs = []  # Final DP5S

        self.output = ""


def ProcessIsomers(dp5Data, Isomers, Settings):

    # extract calculated and experimental shifts and add to dp5Data instance

    # Carbon

    # make sure any shifts with missing peaks are removed from all isomers

    for iso in Isomers:

        dp5Data.Cexp.append([])
        dp5Data.Cshifts.append([])
        dp5Data.Clabels.append([])
        dp5Data.Cinds.append([])
        dp5Data.ConfCshifts.append([])

        a_ind = 0

        exp_inds = []

        for shift, exp, label in zip(iso.Cshifts, iso.Cexp, iso.Clabels):

            if exp != '':

                dp5Data.Cshifts[-1].append(shift)
                dp5Data.Cexp[-1].append(exp)
                dp5Data.Clabels[-1].append(label)
                dp5Data.Cinds[-1].append(int(label[1:]) - 1)
                exp_inds.append(a_ind)

            a_ind += 1

        if len(iso.ConformerCShifts) > 0:

            for conf_shifts in iso.ConformerCShifts:

                dp5Data.ConfCshifts[-1].append( [conf_shifts[e] for e in exp_inds])

    # if the is nmr data append scaled shifts

    if "n" in Settings.Workflow:

        for iso in range(0, len(Isomers)):

            if len(dp5Data.Cshifts[iso]) > 3:

                dp5Data.Cscaled.append(ScaleNMR(dp5Data.Cshifts[iso], dp5Data.Cexp[iso]))

            else:

                dp5Data.Cscaled.append(dp5Data.Cshifts[iso])

            dp5Data.CMAE.append(np.mean(np.abs(np.array(dp5Data.Cscaled[iso]) - np.array(dp5Data.Cexp[iso]))))
            dp5Data.CMax.append(np.max(np.abs(np.array(dp5Data.Cscaled[iso]) - np.array(dp5Data.Cexp[iso]))))

    for iso_ind , iso in enumerate(Isomers):

        InputFile = Path(iso.InputFile)

        # make rdkit mol for each conformer

        dp5Data.Mols.append([])

        if ("m" not in Settings.Workflow) & ("o" not in Settings.Workflow):

            dp5Data.Mols[-1].append(iso.Mols[0])

        else:

            conf_data = []

            if "o" in Settings.Workflow:

                conf_data = iso.DFTConformers

            elif "m" in Settings.Workflow:

                conf_data = iso.Conformers

            for i, geom in enumerate(conf_data):

                m = Chem.MolFromMolFile(str(InputFile) + ".sdf", removeHs=False)

                conf = m.GetConformer(0)

                for j, atom_coords in enumerate(geom):
                    conf.SetAtomPosition(j, Point3D(float(atom_coords[0]), float(atom_coords[1]),
                                                    float(atom_coords[2])))

                dp5Data.Mols[-1].append(m)


    # make pandas df to go into CNN predicting model

    if "n" in Settings.Workflow:

        model = CNN_model.build_model(Settings, "Error")


    else:

        model = CNN_model.build_model(Settings, "Exp")


    for iso in range(len(Isomers)):

        iso_df = []

        i = 0
        for conf in dp5Data.Mols[iso]:
            iso_df += [(i, conf, np.array(dp5Data.Cinds[iso]))]

            i += 1

        iso_df = pd.DataFrame(iso_df, columns=['conf_id', 'Mol', 'atom_index'])

        # use CNN to get reps for each conformer

        if "n" in Settings.Workflow:

            dp5Data.ErrorAtomReps.append(CNN_model.extract_Error_reps(model, iso_df, Settings))

        else:

            dp5Data.ExpAtomReps.append(CNN_model.extract_Exp_reps(model, iso_df, Settings))

    return dp5Data


def kde_probs(Isomers,Settings,DP5type, AtomReps, ConfCshifts,Cexp):

    if DP5type == "Error":

        Error_kernel = pickle.load(open(Path(Settings.ScriptDir) / "pca_10_kde_ERRORrep_Error_kernel.p", "rb"))

        def kde_probfunction(conf_shifts, conf_reps, exp_data):

            # loop through atoms in the test molecule - generate kde for all of them.

            # check if this has been calculated

            n_points = 250

            x = np.linspace(-20, 20, n_points)

            ones_ = np.ones(n_points)

            n_comp = 10

            p_s = []

            scaled_shifts = ScaleNMR(conf_shifts, exp_data)

            scaled_errors = [shift - exp for shift, exp in zip(scaled_shifts, exp_data)]

            for rep, value in zip(conf_reps, scaled_errors):

                # do kde hybrid part here to yield the atomic probability p

                point = np.vstack((rep.reshape(n_comp, 1) * ones_, x))

                pdf = Error_kernel.pdf(point)

                integral_ = np.sum(pdf)

                if integral_ != 0:
                    max_x = x[np.argmax(pdf)]

                    low_point = max(-20, max_x - abs(max_x - value))

                    high_point = min(20, max_x + abs(max_x - value))

                    low_bound = np.argmin(np.abs(x - low_point))

                    high_bound = np.argmin(np.abs(x - high_point))

                    bound_integral = np.sum(pdf[min(low_bound, high_bound): max(low_bound, high_bound)])

                    p_s.append(bound_integral / integral_)

            return p_s

    if DP5type == "Exp":

        Exp_kernel = pickle.load( open(Path(Settings.ScriptDir) / "pca_10_kde_EXP_kernel.p", "rb"))

        def kde_probfunction(_, conf_reps, exp_data):

            # loop through atoms in the test molecule - generate kde for all of them.

            # check if this has been calculated

            n_points = 250

            x = np.linspace(0, 250, n_points)

            ones_ = np.ones(n_points)

            n_comp = 10

            p_s = []

            for rep, value in zip(conf_reps ,exp_data):

                # do kde hybrid part here to yield the atomic probability p

                point = np.vstack((rep.reshape(n_comp, 1) * ones_, x))

                pdf = Exp_kernel.pdf(point)

                integral_ = np.sum(pdf)

                if integral_ != 0:

                    max_x = x[np.argmax(pdf)]

                    low_point = max(0, max_x - abs(max_x - value))

                    high_point = min(250, max_x + abs(max_x - value))

                    low_bound = np.argmin(np.abs(x - low_point))

                    high_bound = np.argmin(np.abs(x - high_point))

                    bound_integral = np.sum(pdf[min(low_bound, high_bound): max(low_bound, high_bound)])

                    p_s.append(bound_integral / integral_)

                else:

                    p_s.append(1)

            return p_s

    #for each atom in the molecule calculate the atomic worry factor

    AtomProbs = [[] for i in range(len(Isomers))]

    for iso in range(len(Isomers)):

        res = [[] for i in AtomReps[iso]]

        AtomProbs[iso] = [[] for i in AtomReps[iso]]

        maxproc = None

        pool = mp.Pool(processes=maxproc)

        if len(ConfCshifts) > 0:

            conf_shifts = ConfCshifts[iso]

        else:

            conf_shifts = [[] for i in AtomReps[iso] ]

        ind1 = 0

        for shifts , conf_reps in zip(conf_shifts , AtomReps[iso])  :


            res[ind1] = pool.apply_async(kde_probfunction,
                                         [conf_shifts[ind1],conf_reps,Cexp[iso]])

            ind1 += 1

        for ind1 in range(len(res)):

            AtomProbs[iso][ind1] = res[ind1].get()

    return AtomProbs


def ScaleNMR(calcShifts, expShifts):

    slope, intercept, r_value, p_value, std_err = stats.linregress(expShifts,
                                                                   calcShifts)
    scaled = [(x - intercept) / slope for x in calcShifts]

    return scaled


def BoltzmannWeight_DP5(Isomers,AtomProbs):

    BAtomProbs = []

    for iso,scaled_probs in zip( Isomers, AtomProbs):

        B_scaled_probs = [0] * len(scaled_probs[0])

        for population, conf_scaled_p in zip(iso.Populations, scaled_probs ):

            for i in range(len(B_scaled_probs)):

                B_scaled_probs[i] += conf_scaled_p[i] * population

        BAtomProbs.append([1 - p for p in B_scaled_probs])

    return BAtomProbs


def Calculate_DP5(BAtomProbs):

    Molecular_probability = []

    for scaled_probs in BAtomProbs:

        #Molecular_probability.append( 1.0 - gmean(scaled_probs))

        Molecular_probability.append(gmean([ p_si for p_si in scaled_probs]))

        #Molecular_probability.append(np.product([1 - p_si for p_si in scaled_probs]))

        #Molecular_probability.append( 1 - np.product([ p_si for p_si in scaled_probs]))

    return Molecular_probability

def Exp_scaling_function(x):

    # Empirical Scaling Function

    return 0.72 / (1 + np.exp(- 17.16382728 * (x - 0.28914008)))


def Rescale_DP5(Mol_probs,BAtomProbs,Settings,DP5type,CMAE):

    if DP5type == "Exp":

        incorrect_kde = pickle.load(open(Path(Settings.ScriptDir) / "Exp_incorrect_kde.p" ,"rb"))

        correct_kde = pickle.load(open(Path(Settings.ScriptDir) / "Exp_correct_kde.p" ,"rb"))

        DP5AtomProbs = [ [] for iso in range(0,len(Mol_probs)) ]

        DP5probs = []

        for iso in range(0, len(Mol_probs)):

            #DP5AtomProbs[iso] = [ Exp_scaling_function(x) for x in BAtomProbs[iso]]

            #DP5probs.append(float( Exp_scaling_function(Mol_probs[iso]) ))

            DP5AtomProbs[iso] = [ x for x in BAtomProbs[iso]]

            DP5probs.append(float( Mol_probs[iso]) )

            #DP5AtomProbs[iso] = [float(correct_kde.pdf(x) / (incorrect_kde.pdf(x) + correct_kde.pdf(x))) for x in
             #                    BAtomProbs[iso]]

            #DP5probs.append(float(correct_kde.pdf(Mol_probs[iso]) / (
             #           incorrect_kde.pdf(Mol_probs[iso]) + correct_kde.pdf(Mol_probs[iso]))))



    elif DP5type == "Error":

        incorrect_kde = pickle.load(open(Path(Settings.ScriptDir) / "Error_incorrect_kde.p" ,"rb"))

        correct_kde = pickle.load(open(Path(Settings.ScriptDir) / "Error_correct_kde.p" ,"rb"))

        DP5AtomProbs = [ [] for iso in range(0,len(Mol_probs)) ]

        DP5probs = []

        print("CMAE" , CMAE)

        for iso in range(0, len(Mol_probs)):

            if CMAE[iso] < 2:

                DP5AtomProbs[iso] = [float(correct_kde.pdf(x) / (incorrect_kde.pdf(x) + correct_kde.pdf(x))) for x in BAtomProbs[iso]]

                DP5probs.append(float(correct_kde.pdf(Mol_probs[iso]) / (incorrect_kde.pdf(Mol_probs[iso]) + correct_kde.pdf(Mol_probs[iso]))))


            else:

                DP5AtomProbs[iso] = BAtomProbs[iso]

                DP5probs.append(Mol_probs[iso])

    else:

        DP5probs = []

        DP5AtomProbs = []

    return DP5probs, DP5AtomProbs


def Pickle_res(dp5Data,Settings):

    data_dic = {"Cshifts": dp5Data.Cshifts,
                "Cscaled": dp5Data.Cscaled,
                "Cexp": dp5Data.Cexp,
                "CMAE": dp5Data.CMAE,
                "CMax": dp5Data.CMax,
                "Clabels": dp5Data.Clabels,
                "Hshifts": dp5Data.Hshifts,
                "Hscaled": dp5Data.Hscaled,
                "Hexp": dp5Data.Hexp,
                "Hlabels": dp5Data.Hlabels,
                "ConfCshifts": dp5Data.ConfCshifts,

                "Mols": dp5Data.Mols,

                "ErrorAtomReps" : dp5Data.ErrorAtomReps,

                "ErrorAtomProbs" : dp5Data.ErrorAtomProbs,

                "B_ErrorAtomProbs" : dp5Data.B_ErrorAtomProbs ,

                "Mol_Error_probs" : dp5Data.Mol_Error_probs ,

                "DP5_Error_probs" : dp5Data.DP5_Error_probs ,

                "ExpAtomReps" : dp5Data.ExpAtomReps ,

                "ExpAtomProbs" : dp5Data.ExpAtomProbs ,

                "B_ExpAtomProbs" : dp5Data.B_ExpAtomProbs ,

                "Mol_Exp_probs" : dp5Data.Mol_Exp_probs ,

                "DP5_Exp_probs" : dp5Data.DP5_Exp_probs

                }

    pickle.dump(data_dic , open(Path(Settings.OutputFolder) / "dp5" / "data_dic.p","wb"))

    return dp5Data


def UnPickle_res(dp5Data,Settings):

    data_dic =  pickle.load(open(Path(Settings.OutputFolder) / "dp5" / "data_dic.p","rb"))

    dp5Data.Cshifts = data_dic["Cshifts"]
    dp5Data.Cscaled = data_dic["Cscaled"]
    dp5Data.Cexp = data_dic["Cexp"]
    dp5Data.CMAE = data_dic["CMAE"]
    dp5Data.CMax = data_dic["CMax"]
    dp5Data.Clabels = data_dic["Clabels"]
    dp5Data.Hshifts = data_dic["Hshifts"]
    dp5Data.Hscaled = data_dic["Hscaled"]
    dp5Data.Hexp = data_dic["Hexp"]
    dp5Data.Hlabels = data_dic["Hlabels"]
    dp5Data.ConfCshifts = data_dic["ConfCshifts"]

    dp5Data.Mols = data_dic["Mols"]
    dp5Data.ErrorAtomReps= data_dic["ErrorAtomReps"]
    dp5Data.ErrorAtomProbs= data_dic["ErrorAtomProbs"]
    dp5Data.B_ErrorAtomProbs= data_dic["B_ErrorAtomProbs"]
    dp5Data.Mol_Error_probs= data_dic["Mol_Error_probs"]
    dp5Data.DP5_Error_probs= data_dic["DP5_Error_probs"]
    dp5Data.ExpAtomReps= data_dic["ExpAtomReps"]
    dp5Data.ExpAtomProbs= data_dic["ExpAtomProbs"]
    dp5Data.B_ExpAtomProbs= data_dic["B_ExpAtomProbs"]
    dp5Data.Mol_Exp_probs= data_dic["Mol_Exp_probs"]
    dp5Data.DP5_Exp_probs = data_dic["DP5_Exp_probs"]

    return dp5Data


def PrintAssignment(dp5Data,DP5AtomProbs,output,Settings):

    isomer = 0

    if "n" in Settings.Workflow:

        for Clabels, Cshifts, Cexp, Cscaled, atom_p in zip(dp5Data.Clabels, dp5Data.Cshifts, dp5Data.Cexp, dp5Data.Cscaled,DP5AtomProbs):

            dp5Data.output += ("\n\nAssigned C shifts for isomer " + str(isomer + 1) + ": ")

            output = PrintNMR(Clabels, Cshifts, Cscaled, Cexp,atom_p, output)

            isomer += 1

    else:

        for Clabels, Cexp,  atom_p in zip(dp5Data.Clabels, dp5Data.Cexp, DP5AtomProbs):

            dp5Data.output += ("\n\nAssigned C shifts for isomer " + str(isomer + 1) + ": ")

            output = PrintNMR_EXP(Clabels, Cexp, atom_p, output)

            isomer += 1

    return output


def PrintNMR(labels, values, scaled, exp,atom_p,output):

    s = np.argsort(values)

    svalues = np.array(values)[s]

    slabels = np.array(labels)[s]

    sscaled = np.array(scaled)[s]

    sexp = np.array(exp)[s]

    atom_p = np.array(atom_p)[s]

    output += ("\nlabel, calc, corrected, exp, error,prob")

    for i in range(len(labels)):

        output += ("\n" + format(slabels[i], "6s") + ' ' + format(svalues[i], "6.2f") + ' '
                           + format(sscaled[i], "6.2f") + ' ' + format(sexp[i], "6.2f") + ' ' +
                           format(sexp[i] - sscaled[i], "6.2f")+ ' ' +
                           format(atom_p[i] , "6.2f"))

    return output


def PrintNMR_EXP(labels, values, atom_p,output):

    s = np.argsort(values)

    svalues = np.array(values)[s]

    slabels = np.array(labels)[s]

    atom_p = np.array(atom_p)[s]

    output += ("\nlabel, exp, error,prob")

    for i in range(len(labels)):

        output += ("\n" + format(slabels[i], "6s") + ' ' + format(svalues[i], "6.2f") + ' '

                            + format(1 - atom_p[i] , "6.2f"))

    return output


def MakeOutput( Isomers, Settings,DP5data,DP5Probs,DP5AtomProbs):

    # add some info about the calculation

    output = ""

    output += Settings.InputFiles[0] + "\n"

    output += "\n" + "Solvent = " + Settings.Solvent

    output += "\n" + "Force Field = " + Settings.ForceField + "\n"

    if 'o' in Settings.Workflow:
        output += "\n" + "DFT optimisation Functional = " + Settings.oFunctional
        output += "\n" + "DFT optimisation Basis = " + Settings.oBasisSet

    if 'e' in Settings.Workflow:
        output += "\n" + "DFT energy Functional = " + Settings.eFunctional
        output += "\n" + "DFT energy Basis = " + Settings.eBasisSet

    if 'n' in Settings.Workflow:
        output += "\n" + "DFT NMR Functional = " + Settings.nFunctional
        output += "\n" + "DFT NMR Basis = " + Settings.nBasisSet

    if Settings.StatsParamFile != "none":
        output += "\n\nStats model = " + Settings.StatsParamFile

    output += "\n\nNumber of isomers = " + str(len(Isomers))

    c = 1

    for i in Isomers:
        output += "\nNumber of conformers for isomer " + str(c) + " = " + str(len(i.Conformers))

        c += 1

    output = PrintAssignment(DP5data,DP5AtomProbs,output,Settings)

    output += ("\n\nResults of DP5: ")

    for i, p in enumerate(DP5Probs):

        output += ("\nIsomer " + str(i + 1) + ": " + format(p * 100, "4.1f") + "%")

    print(output)

    if Settings.OutputFolder == '':

        out = open(str(os.getcwd()) + "/" + str("Output.dp5"), "w+")

    else:

        out = open(os.path.join(Settings.OutputFolder, str("Output.dp5")), "w+")

    out.write(output)

    out.close()

    return output



def predict_Exp_shifts(Settings, Isomers):

    Exp_model = CNN_model.build_Exp_predicting_model(Settings)

    iso_df = []

    i =0

    labels = []

    for iso in Isomers:

        labels.append([])

        InputFile = Path(iso.InputFile)

        # make rdkit mol for each conformer

        m = Chem.MolFromMolFile(str(InputFile) + ".sdf", removeHs=False)

        #else use rdkit optimisation

        AllChem.EmbedMolecule(m, useRandomCoords=True)

        AllChem.MMFFOptimizeMolecule(m)

        Isomers[i].Mols.append(m)

        Isomers[i].Populations.append(1)

        inds = []

        #else use starting point geom.

        for  j , atom in enumerate(m.GetAtoms()):

            if atom.GetAtomicNum() == 6:
                inds.append(j)

                labels[-1].append("C" + str(j + 1))

        iso_df.append((i , m , np.array(inds) ))

        i+=1


    iso_df = pd.DataFrame( iso_df, columns=['conf_id', 'Mol', 'atom_index'])

    shifts = CNN_model.predict_shifts(Exp_model, iso_df ,Settings)

    return shifts,labels,Isomers


from dp5.nmr_processing.nmr_ai import NMRData

from pathlib import Path
import pickle
import logging

from rdkit import Chem

logger = logging.getLogger(__name__)

data = NMRData(["/home/rk582/DP5/test/TPS-1-14/1"],
               "chloroform", Path.cwd())

print(f"NMRData has processed proton FID: {hasattr(data, 'proton_fid')}")

with open('/home/rk582/DP5/test/tps-1-14.preint', 'wb') as f:
    pickle.dump(data, f)



"""
algorithm borrowed from CASCADE paper and is used to run a quick estimate of conformation space
"""

import logging
from pathlib import Path

from rdkit import Chem
from rdkit.Chem import AllChem

from dp5.conformer_search.base_cs_method import BaseConfSearch, ConfData


logger = logging.getLogger(__name__)


class ConfSearchMethod(BaseConfSearch):

    def __init__(self, settings):
        super().__init__(settings)

    def prepare_input(self, inputs):
        # no conversion required
        return inputs

    def __repr__(self) -> str:
        return "ETKDG"

    def _run(self):
        efilter = self.settings["energy_cutoff"] / 4.184
        conf_limit = self.settings["conf_limit"]

        outputs = []
        for input in self.inputs:
            output_exists = Path(f"{input}.confs").exists()
            if not output_exists:
                mol = Chem.MolFromMolFile(f"{input}.sdf", removeHs=False)
                confs, ids, nr = self._genConf(
                    mol, nc=conf_limit, rms=0.125, efilter=efilter, rmspost=0.5
                )
                save = Chem.SDWriter(f"{input}.confs")
                for energy, id in ids:
                    conf = Chem.Mol(mol, confId=id)
                    conf.SetProp("E", f"{energy * 4.184:.2f}")
                    conf.SetProp("_Name", "{}_{}".format(input, id))
                    save.write(conf)
                save.flush()
            outputs.append(f"{input}.confs")
        return outputs

    def _parse_output(self, file):
        out_file = f"{file}.confs"
        atoms = None
        conformers = []
        charge = 0
        energies = []
        suppl = Chem.SDMolSupplier(out_file, removeHs=False)
        for conf in suppl:
            if atoms is None:
                atoms = []
                for atom in conf.GetAtoms():
                    atoms.append(atom.GetSymbol())
                    charge += atom.GetFormalCharge()
            conf3d = conf.GetConformer()
            coords = conf3d.GetPositions().tolist()
            conformers.append(coords)
            energies.append(float(conf.GetProp("E")))
            conf_data = ConfData(atoms, conformers, charge, energies)
        return conf_data

    # algorithm to generate nc conformations
    def _genConf(self, m, nc, rms, efilter, rmspost):
        """
        Generates the conformers.

        Parameters:
        - m(rdkit.Chem.Mol): the molecule on which a search is run
        - nc(int): number of conformers to sample
        - rms(float): RMSD threshold before MMFF optimisation
        - efilter: energy filter (kcal/mol)
        - rmspost: RMSD threshold after MMFF optimisation
        """

        nr = int(AllChem.CalcNumRotatableBonds(m))
        Chem.AssignAtomChiralTagsFromStructure(m, replaceExistingTags=True)
        if not nc:
            nc = min(1000, 3**nr)

        logger.info("ETKDG conformational search")
        logger.info(f"Energy window: {efilter:.2f} kcal/mol")
        logger.debug(f"Will generate up to {nc} conformers")

        if not rms:
            rms = -1
        ids = AllChem.EmbedMultipleConfs(
            m, numConfs=nc, randomSeed=0xF00D, useRandomCoords=True, pruneRmsThresh=rms
        )

        if len(ids) == 0:
            ids = m.AddConformer(m.GetConformer, assignId=True)

        logger.info(f"Generated {len(ids)} conformers")
        diz = []
        diz2 = []
        diz3 = []
        for id in ids:
            prop = AllChem.MMFFGetMoleculeProperties(m, mmffVariant="MMFF94s")
            ff = AllChem.MMFFGetMoleculeForceField(m, prop, confId=id)
            ff.Minimize()
            en = float(ff.CalcEnergy())
            econf = (en, id)
            diz.append(econf)

        if efilter != "Y":
            n, diz2 = self._energy_filter(m, diz, efilter)
        else:
            n = m
            diz2 = diz

        if rmspost != None and n.GetNumConformers() > 1:
            o, diz3 = self._postrmsd(n, diz2, rmspost)
        else:
            o = n
            diz3 = diz2

        return o, diz3, nr

    # filter conformers based on relative energy
    def _energy_filter(self, m, diz, efilter):
        logger.info("Filtering conformers, energy cutoff: %.2f kcal/mol", efilter)
        diz.sort()
        mini = float(diz[0][0])
        sup = mini + efilter
        n = Chem.Mol(m)
        n.RemoveAllConformers()
        n.AddConformer(m.GetConformer(int(diz[0][1])))
        nid = []
        ener = []
        nid.append(int(diz[0][1]))
        ener.append(float(diz[0][0]))
        del diz[0]
        for x, y in diz:
            if x <= sup:
                n.AddConformer(m.GetConformer(int(y)))
                nid.append(int(y))
                ener.append(float(x))
            else:
                break
        diz2 = list(zip(ener, nid))
        logger.info(f"Retained {len(ener)} conformers")
        return n, diz2

    # filter conformers based on geometric RMS
    def _postrmsd(self, n, diz2, rmspost):
        logger.info(f"Filtering conformers, RMS cutoff: {rmspost}")
        diz2.sort(key=lambda x: x[0])
        o = Chem.Mol(n)
        confidlist = [diz2[0][1]]
        enval = [diz2[0][0]]
        nh = Chem.RemoveHs(n)
        del diz2[0]
        for z, w in diz2:
            confid = int(w)
            p = 0
            for conf2id in confidlist:
                # print(confid, conf2id)
                rmsd = AllChem.GetBestRMS(nh, nh, prbId=confid, refId=conf2id)
                if rmsd < rmspost:
                    p = p + 1
                    break
            if p == 0:
                confidlist.append(int(confid))
                enval.append(float(z))
        diz3 = list(zip(enval, confidlist))
        logger.info("Retained %s conformers", len(enval))
        return o, diz3



import numpy as np
import nmrglue as ng

from dp5.nmr_processing.helper_functions import read_jcamp, read_bruker
file_2d = '/home/rk582/DP5/test/isasfid1.dx'
file_1d = '/home/rk582/DP5/test/TESTFID.DX'
bruker_1d = '/home/rk582/DP5/test/TPS-1-14/99'

nucleus, ydata, uc = read_jcamp(file_1d)
assert nucleus == '13C'
assert type(ydata) == np.ndarray
assert type(uc) == ng.fileio.fileiobase.unit_conversion
# Must crash on multidimensional data
try:
    dic, ydata = read_jcamp(file_2d)
    raise AssertionError('nmrglue does not support multidomensional data')
except UserWarning:
    pass
print('JCAMP files read successfully')

nucleus, ydata, uc = read_bruker(bruker_1d)
assert nucleus == '1H'
assert type(ydata) == np.ndarray
assert type(uc) == ng.fileio.fileiobase.unit_conversion
print('Bruker files read sucessfully')



"""Implements the graph convolutional neural network for shift simulation and DP5 probaility estimation"""

from dp5.neural_net.nfp.preprocessing import GraphSequence
from dp5.neural_net.nfp.models import GraphModel
from dp5.neural_net.nfp.layers import (
    Squeeze,
    ReduceBondToAtom,
    GatherAtomToBond,
    ReduceAtomToPro,
)
from pathlib import Path
import pickle
import logging
import warnings
import tempfile
import zipfile
import os

import numpy as np
from numpy.random import seed
import sys
import pandas as pd

# seed(1)
# from tensorflow.random import set_seed
# set_seed(2)

# Define Keras model
import tensorflow as tf
import keras
import keras.backend as K
from keras.layers import Input, Dense, Add, Embedding, Concatenate, Multiply, Add
from keras.models import Sequential, Model, load_model
from keras.callbacks import EarlyStopping
from keras.optimizers import Adam
from keras.optimizers.schedules import ExponentialDecay

from dp5.neural_net import nfp

# essential for models to work!!!
sys.modules["nfp"] = nfp

logger = logging.getLogger(__name__)


def rbf_expansion(distances, mu=0, delta=0.1, kmax=256):
    k = np.arange(0, kmax)
    logits = -((np.atleast_2d(distances).T - (-mu + delta * k)) ** 2) / delta
    return np.exp(logits)


def _atomic_number_tokenizer(atom):
    """suspicious code, likely redundant"""
    return atom.GetNumRadicalElectrons()


def _compute_stacked_offsets(sizes, repeats):
    return np.repeat(np.cumsum(np.hstack([0, sizes[:-1]])), repeats)


class RBFSequence(GraphSequence):
    def process_data(self, batch_data):
        batch_data["distance_rbf"] = rbf_expansion(batch_data["distance"])

        offset = _compute_stacked_offsets(batch_data["n_pro"], batch_data["n_atom"])

        offset = np.where(batch_data["atom_index"] >= 0, offset, 0)
        batch_data["atom_index"] += offset

        del batch_data["n_atom"]
        del batch_data["n_bond"]
        del batch_data["distance"]

        return batch_data


def Mol_iter(dfr):

    for index, r in dfr.iterrows():

        yield (r["Shift"], index)


def Mol_iter2(df):
    for index, r in df.iterrows():
        yield (r["Mol"], r["atom_index"])


def atomic_number_tokenizer(atom):
    return atom.GetAtomicNum()


def message_block(atom_features, atom_state, bond_state, connectivity):
    atom_state = Dense(atom_features, use_bias=False)(atom_state)

    source_atom_gather = GatherAtomToBond(1)
    target_atom_gather = GatherAtomToBond(0)

    source_atom = source_atom_gather([atom_state, connectivity])
    target_atom = target_atom_gather([atom_state, connectivity])

    # Edge update network
    bond_state_message = Concatenate()([source_atom, target_atom, bond_state])
    bond_state_message = Dense(2 * atom_features, activation="softplus")(
        bond_state_message
    )
    bond_state_message = Dense(atom_features)(bond_state_message)

    bond_state_message = Dense(atom_features, activation="softplus")(bond_state_message)
    bond_state_message = Dense(atom_features, activation="softplus")(bond_state_message)
    bond_state = Add()([bond_state_message, bond_state])

    # message function
    messages = Multiply()([source_atom, bond_state])
    messages = ReduceBondToAtom(reducer="sum")([messages, connectivity])

    # state transition function
    messages = Dense(atom_features, activation="softplus")(messages)
    messages = Dense(atom_features)(messages)
    atom_state = Add()([atom_state, messages])

    return atom_state, bond_state


def build_model(model_file):
    """This model returns internal atomic representations from NMR prediction model"""
    # Construct input sequences

    preprocessor = pickle.load(
        open(Path(__file__).parent / "mean_model_preprocessor.p", "rb")
    )

    # Raw (integer) graph inputs
    atom_index = Input(shape=(1,), name="atom_index", dtype="int32")
    atom_types = Input(shape=(1,), name="atom", dtype="int32")
    distance_rbf = Input(shape=(256,), name="distance_rbf", dtype="float32")
    connectivity = Input(shape=(2,), name="connectivity", dtype="int32")
    n_pro = Input(shape=(1,), name="n_pro", dtype="int32")

    squeeze = Squeeze()

    satom_index = squeeze(atom_index)
    satom_types = squeeze(atom_types)
    sn_pro = squeeze(n_pro)
    # Initialize RNN and MessageLayer instances
    atom_features = 256

    # Initialize the atom states
    atom_state = Embedding(
        preprocessor.atom_classes, atom_features, name="atom_embedding"
    )(satom_types)

    bond_state = distance_rbf

    for _ in range(3):
        atom_state, bond_state = message_block(
            atom_features, atom_state, bond_state, connectivity
        )

    atom_state = ReduceAtomToPro(reducer="unsorted_mean")(
        [atom_state, satom_index, sn_pro]
    )

    trans_model = load_model(
        str(Path(__file__).parent / model_file),
        custom_objects={
            "GraphModel": GraphModel,
            "Squeeze": Squeeze,
            "GatherAtomToBond": GatherAtomToBond,
            "ReduceBondToAtom": ReduceBondToAtom,
            "ReduceAtomToPro": ReduceAtomToPro,
        },
    )

    model = GraphModel(
        [atom_index, atom_types, distance_rbf, connectivity, n_pro], [atom_state]
    )

    model.compile()

    # transfer weights from dft model by name:

    model_layers = [n.name for n in model.layers]

    for i, l in enumerate(model.layers):

        l.trainable = True

    for i, trans_layer in enumerate(trans_model.layers):

        if trans_layer.name in model_layers:

            model_layer = model.get_layer(name=trans_layer.name)

            model_layer.set_weights(trans_layer.get_weights())

    return model


def extract_representations(model, test, batch_size):
    with open(Path(__file__).parent / "mean_model_preprocessor.p", "rb") as f:
        preprocessor = pickle.load(f)
    inputs_test = preprocessor.predict(Mol_iter2(test))
    test_sequence = RBFSequence(inputs_test, test.atom_index, batch_size)
    reps_list = []
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        for i in test_sequence:
            yhat = model(i[0])
            indices = i[0]["n_pro"].cumsum()[:-1]
            reps = yhat.numpy()
            reps = np.split(reps, indices)
            # now supports batches!
            reps_list.extend(reps)

    return reps_list
    # finish later


def extract_Error_reps(model, test, Settings):

    batch_size = 1

    preprocessor = pickle.load(
        open(Path(Settings.ScriptDir) / "mean_model_preprocessor.p", "rb")
    )

    inputs_test = preprocessor.predict(Mol_iter2(test))

    test_sequence = RBFSequence(inputs_test, test.atom_index, batch_size)

    pca = pickle.load(
        open(Path(Settings.ScriptDir) / "pca_10_ERRORrep_Error_decomp.p", "rb")
    )

    reps = []

    for i, row in test.iterrows():

        yhat = model(test_sequence[i][0])

        r = [m for m in yhat.numpy()]

        X = pca.transform(r)

        reps.append(X)

        i += 1

    return reps


def extract_Exp_reps(model, test, Settings):

    batch_size = 1

    preprocessor = pickle.load(
        open(Path(Settings.ScriptDir) / "mean_model_preprocessor.p", "rb")
    )

    inputs_test = preprocessor.predict(Mol_iter2(test))

    test_sequence = RBFSequence(inputs_test, test.atom_index, batch_size)

    pca = pickle.load(open(Path(Settings.ScriptDir) / "pca_10_EXP_decomp.p", "rb"))

    reps = []

    for i, row in test.iterrows():

        yhat = model(test_sequence[i][0])

        r = np.array([m for m in yhat.numpy()])

        X = pca.transform(r)

        reps.append(X)

        i += 1

    # do pca on the reps

    return reps


def load_NMR_prediction_model(
    filepath="NMRdb-CASCADEset_Exp_mean_model_atom_features256.hdf5",
):
    """
    Loads NMR predicting model. Make sure your model is in neural_net folder!
    """

    model = load_model(
        str(Path(__file__).parent / filepath),
        custom_objects={
            "GraphModel": GraphModel,
            "Squeeze": Squeeze,
            "GatherAtomToBond": GatherAtomToBond,
            "ReduceBondToAtom": ReduceBondToAtom,
            "ReduceAtomToPro": ReduceAtomToPro,
        },
    )
    return model


def load_quantile_model(
    filepath="CASCADE_quantile_extended.keras",
):
    """
    Loads NMR predicting mode with quantile regression. Make sure your model is in neural_net folder!
    """

    model = load_model(
        str(Path(__file__).parent / filepath),
        custom_objects={
            "GraphModel": GraphModel,
            "Squeeze": Squeeze,
            "GatherAtomToBond": GatherAtomToBond,
            "ReduceBondToAtom": ReduceBondToAtom,
            "ReduceAtomToPro": ReduceAtomToPro,
            "_qloss": QuantileLoss(np.linspace(0.01, 0.99, 99)),
        },
        compile=False,
    )
    # separate compilation required to suppress a warning
    # https://stackoverflow.com/questions/77150716/loading-keras-model-issues-warning-skipping-variable-loading-for-optimizer-ada
    model.compile()
    return model


def get_shifts_and_labels(mols, atomic_symbol, model_path, batch_size=16):
    model = load_NMR_prediction_model(model_path)
    logger.info("Loaded NMR prediction model")
    all_shifts = []

    all_df, all_labels = mols_to_df(mols, atomic_symbol)

    logger.info(f"Ready to predict shifts for {atomic_symbol}")
    all_shifts = predict_shifts(model, all_df, batch_size=batch_size)

    return all_shifts, all_labels


def mols_to_df(mols, atomic_symbol):
    all_labels = []
    all_df = []

    for mol_id, mol in enumerate(mols):
        inds = [
            at.GetIdx() for at in mol[0].GetAtoms() if at.GetSymbol() == atomic_symbol
        ]
        mol_labels = [f"{atomic_symbol}{i+1}" for i in inds]
        all_labels.append(mol_labels)

        for conf_id, conf in enumerate(mol):
            all_df.append((mol_id, conf_id, conf, np.array(inds)))

    all_df = pd.DataFrame(all_df, columns=["mol_id", "conf_id", "Mol", "atom_index"])

    return all_df, all_labels


def predict_shifts(model, test, batch_size=16):
    """Predicts the shifts for all conformers of all molecules.
    - model: prediction_model
    - test: dataframe containing mol_id, conf_id, Mol object and atom indices column
    - batch_size: batch_size
    Returns:
    - list(molecules) of lists(geometries) of lists(shifts)
    """
    preprocessor = pickle.load(
        open(Path(__file__).parent / "mean_model_preprocessor.p", "rb")
    )

    inputs_test = preprocessor.predict(Mol_iter2(test))

    test_sequence = RBFSequence(inputs_test, test.atom_index, batch_size)

    iso_shifts = []
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        for i in test_sequence:
            # returns (n, 1) shape array of n shifts
            yhat = model(i[0])
            indices = i[0]["n_pro"].cumsum()[:-1]
            # must flatten it for shifts
            shifts = yhat.numpy().flatten()
            shifts = np.split(shifts, indices)
            # now supports batches!
            iso_shifts.extend(shifts)

    test["shift_arrays"] = iso_shifts
    all_shifts = [
        np.stack(shifts)
        for i, shifts in test.groupby("mol_id", sort=False)["shift_arrays"]
    ]

    return all_shifts


class PercentileRegressor:
    # create a @classmethod to initialise a model from CASCADE or load an existing one
    def __init__(self, model, quantiles):

        self.quantiles = quantiles
        self.dims = len(self.quantiles)
        self.model = model

    def save(self, archive_path):
        # Use tempfile to create a temporary directory
        with tempfile.TemporaryDirectory() as temp_dir:
            # Save the NumPy array in the temporary directory
            array_path = os.path.join(temp_dir, "array.npy")
            np.save(array_path, self.quantiles)

            # Save the TensorFlow model in the temporary directory
            model_path = os.path.join(temp_dir, "model.keras")
            self.model.save(model_path)

            # Create a ZIP archive containing both the array and the model
            with zipfile.ZipFile(archive_path, "w", zipfile.ZIP_DEFLATED) as zipf:
                # Add the NumPy array to the ZIP archive
                zipf.write(array_path, "array.npy")
                zipf.write(model_path, "model.keras")

    @classmethod
    def load(cls, archive_path):
        # Use tempfile to extract the ZIP archive to a temporary directory
        with tempfile.TemporaryDirectory() as temp_dir:
            # Extract the archive contents into the temporary directory
            with zipfile.ZipFile(archive_path, "r") as zipf:
                zipf.extractall(temp_dir)

            # Load the NumPy array
            array_path = os.path.join(temp_dir, "array.npy")
            arr = np.load(array_path)

            # Load the TensorFlow model
            model_dir = os.path.join(temp_dir, "model.keras")
            model = tf.keras.models.load_model(
                model_dir, custom_objects={"_qloss": QuantileLoss(arr)}
            )

            return cls(model, arr)

    @classmethod
    def from_cascade(
        cls,
        quantiles,
        model_path="NMRdb-CASCADEset_Exp_mean_model_atom_features256.hdf5",
    ):
        dims = len(quantiles)
        full_model = load_NMR_prediction_model(model_path)
        initial_learning_rate = 5e-4
        lr_schedule = ExponentialDecay(
            initial_learning_rate, decay_steps=100000, decay_rate=0.96, staircase=True
        )
        optimizer = Adam(learning_rate=lr_schedule)
        mean_shift = full_model.get_layer(name="atomwise_shift").weights[0][2].numpy()

        input_rep = Input(shape=(256,), name="input_rep")
        x = full_model.get_layer("loc_1")(input_rep)
        x = full_model.get_layer("loc_2")(x)
        x = full_model.get_layer("loc_3")(x)
        x = Dense(dims, name="loc_reduce")(x)
        output = Dense(dims, name="workaround", trainable=False)(x)

        model = Model(input_rep, output)

        w, b = full_model.get_layer("loc_reduce").weights
        w_c = np.broadcast_to(w, shape=(w.shape[0], dims))
        b_c = np.broadcast_to(b, shape=(dims))
        model.get_layer("loc_reduce").set_weights([w_c, b_c])

        w_w, b_w = np.eye(dims), np.broadcast_to(mean_shift, shape=(dims))
        model.get_layer("workaround").set_weights([w_w, b_w])

        model.compile(optimizer=optimizer, loss=QuantileLoss(quantiles))
        class1 = cls(model, quantiles)
        return class1

    def fit(self, *args, **kwargs):
        early_stopping_monitor = EarlyStopping(
            monitor="val_loss",
            min_delta=0,
            patience=10,
            verbose=0,
            mode="auto",
            baseline=None,
            restore_best_weights=True,
        )
        return self.model.fit(*args, **kwargs, callbacks=[early_stopping_monitor])

    def __call__(self, *args, **kwargs):
        return self.predict(*args, **kwargs)

    def predict(self, *args, **kwargs):
        return self.model.predict(*args, **kwargs)


def QuantileLoss(perc, delta=1e-4):
    perc = np.array(perc).reshape(-1)
    perc.sort()
    perc = perc.reshape(1, -1)

    @keras.saving.register_keras_serializable()
    def _qloss(y, pred):
        I = tf.cast(y <= pred, tf.float32)
        d = K.abs(y - pred)
        correction = I * (1 - perc) + (1 - I) * perc
        # huber loss
        huber_loss = K.sum(
            correction * tf.where(d <= delta, 0.5 * d**2 / delta, d - 0.5 * delta), -1
        )
        # order loss
        q_order_loss = K.sum(K.maximum(0.0, pred[:, :-1] - pred[:, 1:] + 1e-6), -1)
        return huber_loss + q_order_loss

    return _qloss


class CASCADE_Quantile:
    def __init__(
        self,
        model,
        quantiles,
    ):
        self.model = model
        self.quantiles = quantiles

    @classmethod
    def from_cascade(
        cls,
        quantiles,
        model_path,
    ):
        """
        Creates a quantile regressor out of CASCADE model. Requires training
        Arguments:
          quantiles: list or numpy array of quantiles to see.
          model_path: path to saved model. Saves all pretrained model weights
        """
        full_model = load_model(
            model_path,
            custom_objects={
                "GraphModel": GraphModel,
                "Squeeze": Squeeze,
                "GatherAtomToBond": GatherAtomToBond,
                "ReduceBondToAtom": ReduceBondToAtom,
                "ReduceAtomToPro": ReduceAtomToPro,
            },
        )
        *_, penultimate_name, final_name = [
            layer.name
            for layer in full_model.layers
            if layer.__class__.__name__ == "Dense"
        ]

        *_, reduce_layer_name = [
            layer.name
            for layer in full_model.layers
            if layer.__class__.__name__ == "ReduceAtomToPro"
        ]
        quantiles = np.array(quantiles)
        quantiles = np.sort(quantiles)
        dims = len(quantiles)

        rep, atomwise_shift = (
            full_model.get_layer(name=penultimate_name).output,
            full_model.get_layer(name=reduce_layer_name).output,
        )
        rep = Dense(dims, name=final_name)(rep)
        output = Add(name="final_layer")([rep, atomwise_shift])

        model = GraphModel(inputs=full_model.input, outputs=output)

        w, b = full_model.get_layer(final_name).weights
        w_c = np.broadcast_to(w, shape=(w.shape[0], dims))
        b_c = np.broadcast_to(b, shape=(dims))
        model.get_layer(final_name).set_weights([w_c, b_c])

        initial_learning_rate = 5e-4
        lr_schedule = ExponentialDecay(
            initial_learning_rate, decay_steps=100000, decay_rate=0.96, staircase=True
        )
        optimizer = Adam(learning_rate=lr_schedule)

        model.compile(optimizer=optimizer, loss=QuantileLoss(quantiles))

        ready = cls(model, quantiles)

        return ready

    def save(self, archive_path):
        # Use tempfile to create a temporary directory
        with tempfile.TemporaryDirectory() as temp_dir:
            # Save the NumPy array in the temporary directory
            array_path = os.path.join(temp_dir, "array.npy")
            np.save(array_path, self.quantiles)

            # Save the TensorFlow model in the temporary directory
            model_path = os.path.join(temp_dir, "model.keras")
            self.model.save(model_path)

            # Create a ZIP archive containing both the array and the model
            with zipfile.ZipFile(archive_path, "w", zipfile.ZIP_DEFLATED) as zipf:
                # Add the NumPy array to the ZIP archive
                zipf.write(array_path, "array.npy")
                zipf.write(model_path, "model.keras")

    @classmethod
    def load(cls, archive_path):
        # Use tempfile to extract the ZIP archive to a temporary directory
        with tempfile.TemporaryDirectory() as temp_dir:
            # Extract the archive contents into the temporary directory
            with zipfile.ZipFile(archive_path, "r") as zipf:
                zipf.extractall(temp_dir)

            # Load the NumPy array
            array_path = os.path.join(temp_dir, "array.npy")
            arr = np.load(array_path)

            # Load the TensorFlow model
            model_dir = os.path.join(temp_dir, "model.keras")
            model = tf.keras.models.load_model(
                model_dir,
                custom_objects={
                    "GraphModel": GraphModel,
                    "Squeeze": Squeeze,
                    "GatherAtomToBond": GatherAtomToBond,
                    "ReduceBondToAtom": ReduceBondToAtom,
                    "ReduceAtomToPro": ReduceAtomToPro,
                    "_qloss": QuantileLoss(arr),
                },
                compile=False,
            )
            initial_learning_rate = 5e-4
            lr_schedule = ExponentialDecay(
                initial_learning_rate,
                decay_steps=100000,
                decay_rate=0.96,
                staircase=True,
            )
            optimizer = Adam(learning_rate=lr_schedule)

            model.compile(optimizer=optimizer, loss=QuantileLoss(arr))
            return cls(model, arr)

    def fit(self, *args, **kwargs):
        return self.model.fit(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        return self.predict(*args, **kwargs)

    def predict(self, *args, **kwargs):
        return self.model(*args, **kwargs)



"""expands NMRFile for input. Functionality is moved into NMRData"""

from pathlib import Path


def find_nmr_files(nmr_path: str) -> list[Path]:
    """
    takes path for nmr file, returns the path of relevant files

    arguments:
    - nmr_path: relative path to NMRFile

    returns:
    list of paths to files
    """

    _path = (Path(nmr_path)).resolve()
    nmr_paths = []

    if _path.is_dir():
        c_path = ""
        h_path = ""
        for file in _path.iterdir():
            if file.name.lower() == "carbon" or file.name.lower() == "carbon.dx":
                c_path = file
            elif file.name.lower() == "proton" or file.name.lower() == "proton.dx":
                h_path = file
            if c_path and h_path:
                break
        nmr_paths = [c_path, h_path]

    else:
        nmr_paths.append(_path)

    return nmr_paths



__all__ = ["setup_logger"]
import sys
import logging
import os
import csv
import io
from logging.config import dictConfig, fileConfig

def setup_logger(
    name: str = None,
    config: dict = None,
    filename: str = None,
    formatter=None,
    stream=sys.stderr,
    cfg_filename: str = None,
    propagate: bool = True,
    level=logging.INFO,
    debug=False,
):
    """Setup a logging facility.

    :param name: name of the logger, root if empty or None
    :param config: dictionary configuration
    :param filename: optional filename for logging output
    :param formatter: a logging formatter
    :param stream: the output stream
    :param cfg_filename: filename of a logger configuration file
    :param propagate: whether to propagate to higher level loggers
    :param level: logging level
    :param debug: set special format for debugging
    :returns: the newly set up logger
    """

    logging.captureWarnings(True)

    logger = logging.getLogger(name)
    logger.setLevel(level)

    for handler in logger.handlers[:]:
        logger.removeHandler(handler)

    if config is not None:
        dictConfig(config)
        return

    if cfg_filename is not None:
        fileConfig(cfg_filename)
        return

    if filename:
        handler = logging.FileHandler(filename, mode="w+")
    else:
        handler = logging.StreamHandler(stream)

    handler.setLevel(level)

    if debug:
        log_format = "%(asctime)s %(module)s.%(funcName)s +%(lineno)s: %(levelname)-4s %(message)s"
    else:
        log_format = "%(asctime)s <%(levelname)-4.4s> %(message)s"

    if not formatter:
        formatter = logging.Formatter(
            fmt=log_format,
            datefmt="%H:%M:%S",
        )

    handler.setFormatter(formatter)

    logger.addHandler(handler)
    logger.propagate = propagate

    return logger